from classes.User import User
from classes.Transaction import Tx
from classes.TxBlock import TxBlock

from helpers import DatabaseHelper as databaseHelper
from helpers import UtilityHelper as utilityHelper
from helpers import BlockHelper as blockHelper
from helpers import TransactionHelper as transactionHelper
from helpers import TaskHelper as taskHelper

import socket
import pickle
import select

import os
from dotenv import load_dotenv
load_dotenv()

BUFFER_SIZE = 1024

def newServerSocket(ip_addr, port):
    server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    server_socket.bind((ip_addr, port))
    server_socket.listen()
    return server_socket


def recvObj(socket):
    inputs, outputs, errs = select.select([socket], [], [socket], 6)
    if socket in inputs:
        connected_socket, addr = socket.accept()
        all_data = b''
        while True:
            data = connected_socket.recv(BUFFER_SIZE)
            if not data:
                break
            all_data = all_data + data
        return pickle.loads(all_data)
    return None


def sendObj(ip_addr, blk, port):
    try:
        if type(blk) == TxBlock:
            blk.previousBlock = None
        soc = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        soc.connect((ip_addr, port))
        data = pickle.dumps(blk)
        soc.send(data)
        soc.close()
        return True
    except:
        print("Could not connect to peer")
        utilityHelper.close()
        return False

def connection(ip, port, db):
    server_socket = newServerSocket(ip, port)
    while(True):

        item = recvObj(server_socket)
        if not item:
            continue

        if type(item) == Tx:
            print()
            print(f"{utilityHelper.blinkMessage('Received transaction from peer')}")
            print(">>: ")

            if item.is_valid():
                if item.type == 0:
                    if item.outputs[0][1] > taskHelper.getBalanceWithOutgoingPoolByPublicKey(item.inputs[0][0], taskHelper.getBalanceByPublicKey(item.inputs[0][0])):
                        print(f"{utilityHelper.errorMessage('Received transaction is not valid')}")
                        sendObj(os.getenv("PEER_IP"), ["INVALID TRANSACTION",item], int(os.getenv("PEER_PORT")))
                        continue
                utilityHelper.addFile("../data/transaction_pool.dat", item)
                continue
            
            print(f"{utilityHelper.errorMessage('Received transaction is not valid')}")
            sendObj(os.getenv("PEER_IP"), ["INVALID TRANSACTION",item], int(os.getenv("PEER_PORT")))

        if type(item) == TxBlock:
            print()
            print(f"{utilityHelper.blinkMessage('Received block from peer')}")
            print(">>: ")

            valid, exit, block = blockHelper.socketBlock(item)

            if valid:
                sendObj(os.getenv("PEER_IP"), ["VALIDATE BLOCK",block], int(os.getenv("PEER_PORT")))
                continue

            if not valid:
                print(f"{utilityHelper.errorMessage('Received block is not valid')}")
                sendObj(os.getenv("PEER_IP"), ["INVALID BLOCK",block], int(os.getenv("PEER_PORT")))
                continue         

            if exit:
                print()
                utilityHelper.closeConnection(os.getenv("PEER_IP"), int(os.getenv("PEER_PORT")))
                continue
            continue

        # DATABASE STUFF
        if type(item) == list:
            match item[0]:

                case "USER CREATE":
                    print()
                    print(f"{utilityHelper.blinkMessage('Received user from peer')}")
                    print(">>: ")
                    db.createUser(item[1], item[2], item[3], item[4])
                    continue

                case "USER DELETE":
                    print()
                    print(f"{utilityHelper.blinkMessage('Peer deleted user')}")
                    print(">>: ")
                    db.deleteUser(item[1])
                    continue

                case "USER CHANGE PASSWORD":
                    db.changePassword(item[1], item[2], item[3])
                    continue

                case "USER CHANGE USERNAME":
                    db.changeUsername(item[1], item[2])
                    continue

                case "CREATE LOG":
                    db.createLog(item[1], item[2], item[3])
                    continue

                case "UPDATE LOG STATUS":
                    db.updateLogStatus(item[1])
                    continue

                case "INVALID TRANSACTION":
                    print()
                    print(f"{utilityHelper.errorMessage('Last send transaction is not valid, removing from pool...')}")
                    print(">>: ")
                    transactionHelper.removeTransaction(item[1])
                    continue

                case "INVALID BLOCK":
                    print()
                    print(f"{utilityHelper.errorMessage('Last send block is not valid, removing from ledger...')}")
                    blockHelper.removeInvalid(item[1])
                    print(">>: ")
                    continue

                case "VALIDATE BLOCK":
                    blockHelper.updateBlockValidation(item[1])
                    continue

        if type(item) == str:
            match item:
                case "EXIT":
                    utilityHelper.close()
                    continue
                case "TEST":
                    continue